/*************************************************************************************/
/* Workshop de
 * Supercomputación y Cálculo Científico, SC3-2014
 * SC3UIS con Academia NVIDIA CUDA
 
  Este codigo presenta la obtencion de indices en la API CUDA C. 
  se hace uso de transferencias de memoria global y ejecucion, 
  multithread y multiblock para introducir los conceptos:
  
    * Manejo de memoria en host y device,
    * Copiado y transferencia de variables entre host y device 
    * Invocacion de kernel.
  
 * Creado por: William J Trigos, Ing. Electrónico, UIS(2013) 
 * Fecha: 20/04/2014 12:49 PM, (Bucaramanga, Colombia)
 */
/*************************************************************************************/

#include <stdlib.h>
#include <stdio.h>

// kernel que obtiene los indices de cuda 
__global__ void kernel(int *arrayBl, int *arrayTh){

  // obteniendo los indices 1D para las grid_x y grid_y sentencias reservadas blockId, threadId, blockDim
  int index_x = blockIdx.x * blockDim.x + threadIdx.x;
  int index_y = blockIdx.y * blockDim.y + threadIdx.y;

  // deficion de un arreglo lineal para convertir los indices idx e idy
  // en un solo arreglo 1D
  int grid_width = gridDim.x * blockDim.x;
  int index = index_y * grid_width + index_x;

  // usando el mismo concepto de mapeo 1D para la grid_x y grid_y, 
  // para crear un indice para reconocer los bloques 
  int result = blockIdx.y * gridDim.x + blockIdx.x;

  // write out the result
  arrayBl[index] = result;
  arrayTh[index] = index;
}

//  Programa principal
int main(void)
{
  int num_elements_x = 8;         //16;
  int num_elements_y = 8;         //16;

  int num_bytes = num_elements_x * num_elements_y * sizeof(int);

  int *device_arrayBl = 0;
  int *device_arrayTh = 0;
  int *host_arrayBl = 0;
  int *host_arrayTh = 0;

  // // reservando memoria sobre la CPU con el espacio necesario
  host_arrayBl = (int*)malloc(num_bytes);
  host_arrayTh = (int*)malloc(num_bytes);

  // reservando memoria sobre la GPU con el espacio necesario
  cudaMalloc((void**)&device_arrayBl, num_bytes);
  cudaMalloc((void**)&device_arrayTh, num_bytes);

  // creando bloques de hilos de dimensiones block_size.y * block_size.x
  dim3 block_size;
  block_size.x = 4;               //4;
  block_size.y = 4;               //4;

  // configurando una grid 2D, para lanzamiento de trabajo paralelo sobre los hilos
  dim3 grid_size;
  grid_size.x = num_elements_x / block_size.x;
  grid_size.y = num_elements_y / block_size.y;

  // los parametros grid_size & block_size son pasados para el lanzamiento del kernel cuda
  // e iniciar asi el trabajo paralelo
  kernel <<<grid_size, block_size>>> (device_arrayBl, device_arrayTh);

  // descargar el valor de los indices almacenados en los vectores de la GPU
  // hacia los vectores en la CPU 
  cudaMemcpy(host_arrayBl, device_arrayBl, num_bytes, cudaMemcpyDeviceToHost);
  cudaMemcpy(host_arrayTh, device_arrayTh, num_bytes, cudaMemcpyDeviceToHost);

  // Hacer la impresion de los resultados desde la CPU
  for(int row = 0; row < num_elements_y; ++row){
    for(int col = 0; col < num_elements_x; ++col){
      //printf("%2d |", host_arrayBl[row * num_elements_x + col]);	// Imprime en una cuadrilla el bloque al que pertenece cada Thread
      //printf(" [%2d]	|", host_arrayTh[row * num_elements_x + col]); 	// Imprime los vecinos del thread (TOPOLOGIA CENTRADA - izq, arriba, centro, abajo, derecha)
      printf("[%1d] %2d	|", host_arrayBl[row * num_elements_x + col], host_arrayTh[row * num_elements_x + col]); // Imprime Bloque y thread Respectivo
    }
    printf("\n");
  }
  printf("\n");

  // Liberar memoria en CPU
  free(host_arrayBl);
  free(host_arrayTh);
  // Liberar memoria en GPU
  cudaFree(device_arrayBl);
  cudaFree(device_arrayTh);
}
